import numpy as np
import pathlib
import ase
import sys
from IPython.display import HTML, display
from dataclasses import dataclass
from typing import Sequence, Dict, Optional, List
import matplotlib.pyplot as plt


@dataclass
class Contour:
    name: str  #: JDFTx variable name
    level: float  #: Contour level
    color: Sequence[float]  #: RGBA color + opacity for contour
    field: Optional[np.ndarray] = None  #: Corresponding scalar field data

def read_vibrations(out_file: str) -> List[float]:
    """ 
    Read vibrational modes from JDFTx calculation and return a list of frequencies
    in cm^-1
    """
    freqs = []
    for iLine, line in enumerate(open(out_file)):
        if line.startswith("Frequency"):
            freq = line.split("[")[-1].split()[0]
            freqs.append(freq)
    return freqs
 
def read_energy(out_file: str) -> Dict[str, float]:
    Es = []
    E_str = "F ="
    for iLine, line in enumerate(open(out_file)):
        if "target-mu" in line:
            E_str = "G ="
        if line.strip().startswith(E_str):
            E = line.split()[-1]
            Es.append(E)
    return float(Es[-1])

def plot_pathway(energies:List[float], labels=[]) -> None:
    ax = plt.gca()
    label_locs = []
    for ie, energy in enumerate(energies):
        if ie == 0:
            ax.hlines(energy, 0, 1, color="k", label=labels[ie])
            ax.plot([1,2], [energy, energies[ie+1]], linestyle="--", color="k")
            label_locs.append(0.5)
        elif ie == len(energies) - 1:
            x_start = ie *2
            x_fin = x_start + 1
            label_locs.append(x_start + 0.5)
            ax.hlines(energy, x_start, x_fin, color="k", label=labels[ie])
        else:
            x_start = ie *2
            x_fin = x_start + 1
            label_locs.append(x_start + 0.5)
            ax.hlines(energy, x_start, x_fin, color="k", label=labels[ie])
            ax.plot([x_start+1, x_fin+1], [energy, energies[ie+1]], linestyle="--", color="k")
    ax.set_xticks(label_locs)
    ax.set_xticklabels(labels)
            
    
def read_JDFTx(out_file: str, contours: Optional[List[Contour]] = None) -> ase.Atoms:
    """Read structure in ASE form from JDFTx output file `out_file`.
    Optionally, retrieve scalar fields for each variable named within `contours`."""
    init_done = False
    R = np.zeros((3,3))
    ref_line = -10
    Rdone = False
    Sdone = False
    ion_pos_started = False
    ion_names = []
    ion_pos = []
    for iLine, line in enumerate(open(out_file)):
        if line.startswith("Executable "):
            tokens = line.split()
            in_file = None
            for i_token, token in enumerate(tokens):
                if token in {"-i", "--input-file"}:
                    in_file = tokens[i_token+1]
        if line.startswith('Initialization completed'):
            init_done = True
        if line.find('Initializing the Grid') >= 0:
            ref_line = iLine
        if not Rdone:
            rowNum = iLine - (ref_line+2)
            if rowNum>=0 and rowNum<3:
                R[rowNum,:] = np.array([ float(x) for x in line.split()[1:-1] ])
            if rowNum==3:
                Rdone = True
        if ion_pos_started:
            tokens = line.split()
            if len(tokens) and tokens[0] == 'ion':
                ion_names.append(tokens[1])
                ion_pos.append([float(tokens[2]),float(tokens[3]),float(tokens[4])])
            else:
                break
        if line.find('# Ionic positions in') >= 0:
            coords = line.split()[4]
            ion_pos_started = True
        if (not Sdone) and line.startswith("Chosen fftbox size"):
            tokens = line.split()
            S = np.array([ int(x) for x in line.split()[-4:-1] ])
            Sdone = True
        if line.startswith("dump-name"):
            dump_name = line.split()[-1]
            assert dump_name.find("$STAMP") < 0
            assert dump_name.find("$ITER") < 0
            if dump_name.find("$INPUT") >= 0:
                # Replace with input file prefix:
                assert in_file is not None
                in_prefix = pathlib.Path(in_file).stem
                dump_name = dump_name.replace("$INPUT", in_prefix)

    # --- ensure ionpos in Cartesian coordinates for ASE:
    ion_pos = np.array(ion_pos)
    if coords.lower() != "cartesian":
        ion_pos = ion_pos @ R.T

    # Read any requested scalar fields:
    if contours is not None:
        # Read unique set of scalar fields:
        names = set(contour.name for contour in contours)
        fields = {
            name: np.fromfile(dump_name.replace("$VAR", name)).reshape(S)
            for name in names
        }
        # Replace within contours:
        for contour in contours:
            contour.field = fields[contour.name]

    # Return as ASE Atoms object:
    Angstrom = 1/0.5291772
    return ase.Atoms(ion_names, positions=ion_pos/Angstrom, cell=R.T/Angstrom, pbc=True)


def x3d_view(
    atoms: ase.Atoms,
    repeat: Sequence[int] = (1,1,1),
    contours: Optional[List[Contour]] = None,
) -> HTML:
    # Start HTML page and scene:
    from io import StringIO
    output = StringIO()
    output.write(
        """
        <html><head><title>ASE atomic visualization</title>
        <link rel="stylesheet" type="text/css" href="https://www.x3dom.org/release/x3dom.css"></link>
        <script type="text/javascript" src="https://www.x3dom.org/release/x3dom.js"></script>
        </head>
        <body><X3D><Scene>
        """
    )
    
    # Create shape template for each atom type:
    from ase.data import covalent_radii
    from ase.data.colors import jmol_colors
    atom_numbers = {atom.number for atom in atoms}
    atom_shapes = {}
    for atom_number in atom_numbers:
        r, g, b = jmol_colors[atom_number]
        radius = covalent_radii[atom_number]
        atom_shapes[atom_number] = (
            f"""
            <Shape>
            <Appearance>
            <Material diffuseColor="{r:.3f} {g:.3f} {b:.3f}" specularColor="0.5 0.5 0.5"/>
            </Appearance>
            <Sphere radius="{radius:.2f}"/>
            </Shape>
            """
        )
    # Offsets for repetitions:
    RbasisT = np.array(atoms.get_cell())
    offsets1d = [np.arange(repeat_i) - ((repeat_i - 1) // 2) for repeat_i in repeat]
    offsets = np.array(np.meshgrid(*offsets1d, indexing='ij')).reshape(3, -1).T
    offsets_cart = offsets @ RbasisT
    # Write atoms:
    pos_arr = []
    for offset_cart in offsets_cart:
        for atom in atoms:
            pos = atom.position + offset_cart
            x, y, z = pos
            output.write(
                f"""
                <Transform translation="{x:.2f} {y:.2f} {z:.2f}">
                {atom_shapes[atom.number]}
                </Transform>
                """
            )
            pos_arr.append(pos)
    # Write unit cell:
    verts = np.array(np.meshgrid([0,1], [0,1], [0,1], indexing='ij')).reshape(3, -1).T
    verts_cart = verts @ RbasisT
    output.write(
        f"""
        <Shape>
        <Appearance><Material diffuseColor='0 0 0' specularColor='0 0 0'/></Appearance>
        <IndexedLineSet coordIndex='0 1 3 2 0 -1 4 5 7 6 4 -1 0 4 -1 1 5 -1 2 6 -1 3 7 -1'>
        <Coordinate point='{np.array2string(verts_cart.flatten())[1:-1]}'/>
        </IndexedLineSet>
        </Shape>
        """
    )
    
    # Compute bounding box:
    pos_arr = np.vstack((np.array(pos_arr), verts_cart))
    bbox_min = pos_arr.min(axis=0)
    bbox_max = pos_arr.max(axis=0)

    # Isosurfaces:
    if contours is not None:
        from skimage import measure
        np.set_printoptions(linewidth=100, precision=5, suppress=True, threshold=sys.maxsize)
        for contour in contours:
            assert contour.field is not None
            field_min, field_max = contour.field.min(), contour.field.max()
            if not (field_min < contour.level < field_max):
                raise ValueError(
                    f"Contour level must be between {field_min} and {field_max}"
                    f"for '{contour.name}'"
                )
            verts_mesh, faces, _, _ = measure.marching_cubes(
                contour.field, contour.level, allow_degenerate=False
            )
            dx = 1.0 / np.array(contour.field.shape)  # mesh -> lattice factor
            verts_cart = (verts_mesh * dx[None]) @ RbasisT
            # Convert to X3D mesh:
            r, g, b, alpha = contour.color
            index_str = np.array2string(
                np.hstack((faces, np.full((faces.shape[0],1), -1))).flatten()
            )[1:-1]
            output.write(
                f"""
                <Shape>
                <Appearance>
                <Material diffuseColor="{r:.3f} {g:.3f} {b:.3f}"
                    specularColor="0.5 0.5 0.5" transparency='{1-alpha}'/>
                </Appearance>
                <IndexedFaceSet solid='false' coordIndex='{index_str}'>
                <Coordinate point='{np.array2string(verts_cart.flatten())[1:-1]}'/>
                </IndexedFaceSet>
                </Shape>
                """
            )

    # Set viewpoint based on bounding box:
    x0, y0, z0 = 0.5 * (bbox_min + bbox_max)
    view_distance = np.linalg.norm(bbox_max - bbox_min) * 1.5
    output.write(
        f"""
        <ViewPoint
        position='{x0:.1f} {y0:.1f} {z0+view_distance}'
        centerOfRotation='{x0:.1f} {y0:.1f} {z0:.1f}'/>
        """
    )

    # Close scene and HTML page:
    output.write("</Scene></X3D></body></html>")
    return display(HTML(output.getvalue()))

